#!/usr/bin/env python3

"""
Unittests for Botan Python scripts. Those tests only need to pass un UNIX-like
operating systems.

Requires Python 3.

(C) 2017 Simon Warta (Kullo GmbH)

Botan is released under the Simplified BSD License (see license.txt)
"""

import os
import sys
import unittest

sys.path.append("../..") # Botan repo root
from install import prepend_destdir # pylint: disable=wrong-import-position
from install import PrependDestdirError # pylint: disable=wrong-import-position


class PrependDestdir(unittest.TestCase):
    def test_absolute_destdir(self):
        os.environ["DESTDIR"] = "/"
        self.assertEqual(prepend_destdir("/home/me"), "/home/me")
        self.assertEqual(prepend_destdir("/home/me/"), "/home/me")
        self.assertEqual(prepend_destdir("/home/me/../me2"), "/home/me2")

        os.environ["DESTDIR"] = "/opt"
        self.assertEqual(prepend_destdir("/home/me"), "/opt/home/me")
        self.assertEqual(prepend_destdir("/home/me/"), "/opt/home/me")
        self.assertEqual(prepend_destdir("/home/me/../me2"), "/opt/home/me2")

    def test_relative_destdir(self):
        os.environ["DESTDIR"] = "."
        self.assertEqual(prepend_destdir("/home/me"), "./home/me")
        self.assertEqual(prepend_destdir("/home/me/"), "./home/me")
        self.assertEqual(prepend_destdir("/home/me/../me2"), "./home/me2")

        os.environ["DESTDIR"] = "bar"
        self.assertEqual(prepend_destdir("/home/me"), "bar/home/me")
        self.assertEqual(prepend_destdir("/home/me/"), "bar/home/me")
        self.assertEqual(prepend_destdir("/home/me/../me2"), "bar/home/me2")

    def test_relative(self):
        # No destdir set
        os.environ["DESTDIR"] = ""
        self.assertEqual(prepend_destdir("foo"), "foo")
        self.assertEqual(prepend_destdir("../foo"), "../foo")

        # Destdir set
        os.environ["DESTDIR"] = "/opt"
        with self.assertRaises(PrependDestdirError):
            prepend_destdir("foo")
        with self.assertRaises(PrependDestdirError):
            prepend_destdir("../foo")

    def test_escaping(self):
        os.environ["DESTDIR"] = "/opt"
        with self.assertRaises(PrependDestdirError):
            prepend_destdir("/foo/../..")


if __name__ == '__main__':
    unittest.TestCase.longMessage = True
    unittest.main()



# -* coding: utf-8 -*-
# Sphinx configuration file

import sys
import re

#import sphinx

def check_for_tag(tag):
    # Nasty hack :(
    try:
        opt_t = sys.argv.index('-t')
        opt_tag = sys.argv.index(tag)
        return opt_t + 1 == opt_tag
    except ValueError:
        return False


def parse_version_file(version_path):
    version_file = open(version_path)
    key_and_val = re.compile(r"([a-z_]+) = ([a-zA-Z0-9:\-\']+)")

    results = {}
    for line in version_file.readlines():
        if not line or line[0] == '#':
            continue
        match = key_and_val.match(line)
        if match:
            key = match.group(1)
            val = match.group(2)

            if val == 'None':
                val = None
            elif val.startswith("'") and val.endswith("'"):
                val = val[1:len(val)-1]
            else:
                val = int(val)

            results[key] = val
    return results

version_info = parse_version_file('../../build-data/version.txt')

version_major = version_info['release_major']
version_minor = version_info['release_minor']
version_patch = version_info['release_patch']
version_suffix = version_info['release_suffix']

is_website_build = check_for_tag('website')

needs_sphinx = '1.2'

templates_path = ['templates']

source_suffix = '.rst'

source_encoding = 'utf-8-sig'

master_doc = 'contents'

project = u'botan'
copyright = u'2000-2023, The Botan Authors'

version = '%d.%d' % (version_major, version_minor)
release = '%d.%d.%d%s' % (version_major, version_minor, version_patch, version_suffix)

#today = ''
today_fmt = '%Y-%m-%d'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

highlight_language = 'cpp'

# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

try:
    # On Arch this is python-sphinx-furo
    import furo # noqa: F401
    html_theme = "furo"

    # Add a small edit button to each document to allow visitors to easily
    # propose changes to that document using the repository’s source control system.
    html_theme_options = {
        'source_repository': 'https://github.com/randombit/botan/',
        'source_branch': 'master',
        'source_directory': 'doc/',
    }
except ImportError:
    print("Could not import furo theme; falling back to agago")
    html_theme = 'agogo'
    html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'Botan'

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%Y-%m-%d'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
if is_website_build:
    html_use_opensearch = 'https://botan.randombit.net/'
else:
    html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'botandoc'

# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).

authors = u'The Botan Authors'
latex_documents = [
    ('contents', 'botan.tex', u'Botan Reference Guide', authors, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
latex_show_pagerefs = False

# If true, show URL addresses after external links.
latex_show_urls = 'inline'

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = False

latex_elements = {
    'printindex': '\\footnotesize\\raggedright\\printindex'
}

# Give all sections a label, so we can reference them
extensions = [
    "sphinx.ext.autosectionlabel",
]

# Make sure the target is unique
autosectionlabel_prefix_document = True

# -- Options for texinfo output --------------------------------------------------
authors = 'The Botan Authors'

# Show URL addresses after external links, options are 'inline', 'footnote' and 'no'
texinfo_show_urls = 'inline'



# This test data is generated from the *_DSA implementations of https://github.com/mjosaarinen/py-acvp-pqc/tree/main
# and using the drbg.aes256_ctr_drbg module from https://github.com/giacomopope/dilithium-py
#
# (C) 2024 Jack Lloyd
# (C) 2024 Amos Treiber, René Meusel - Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import binascii
import hashlib
import os
import re
import sys
from dilithium_py.drbg.aes256_ctr_drbg import AES256_CTR_DRBG
from fips204 import ML_DSA, ML_DSA_PARAM
from fips205 import SLH_DSA_PARAMS

random_bytes = os.urandom

# "ML-DSA-44" -> "ml-dsa-4x4"
def transform_mldsa_string(s):
    s = s.lower()
    transformed = re.sub(r'(\d)(\d)$', r'\1x\2', s)
    return transformed

def create_mldsa_ini_label(s, det):
    algo = re.sub(r'(\d)(\d)$', r'\1x\2', s)
    return f"[{algo}_{det}]"

def sha3_256(v):
    h = hashlib.sha3_256()
    h.update(v)
    return h.digest()


def sha256(v):
    h = hashlib.sha256()
    h.update(v)
    return h.digest()

def sign(sign_internal_func, sk_bytes, m, random_bytes, n, ctx=b"", deterministic=False):
     if len(ctx) > 255:
         raise ValueError(
             f"ctx bytes must have length at most 255, ctx has length {len(ctx) = }"
         )
     if deterministic:
         rnd = None
     else:
         rnd = random_bytes(n)   #drbg.random_bytes(n)
     # Format the message using the context
     m_prime = bytes([0]) + bytes([len(ctx)]) + ctx + m
     # Compute the signature of m_prime
     sig_bytes = sign_internal_func(m_prime, sk_bytes, rnd)
     return sig_bytes

def main(args = None):
    msg = binascii.unhexlify("D81C4D8D734FCBFBEADE3D3F8A039FAA2A2C9957E835AD55B22E75BF57BB556AC8")
    if args is None:
        args = sys.argv

    if len(args) != 2 or (args[1] != "ml_dsa" and args[1] != "slh_dsa"):
        print("Usage: gen_pqc_dsa_kats.py <ml_dsa|slh_dsa>")
        return 1

    algo = args[1]

    # Initialize DRBG with the magic value
    entropy = binascii.unhexlify("60496cd0a12512800a79161189b055ac3996ad24e578d3c5fc57c1e60fa2eb4e550d08e51e9db7b67f1a616681d9182d")
    drbg = AES256_CTR_DRBG(seed=entropy)

    if algo == "slh_dsa":
        output = open('src/tests/data/pubkey/slh_dsa.vec', 'w')

        print("# See src/scripts/dev_tools/gen_pqc_dsa_kats.py\n", file=output)

        for param, alg in SLH_DSA_PARAMS.items():
            hash_fn = sha256 if "sha2" in param.lower() else sha3_256

            seed = drbg.random_bytes(len(entropy))

            drbg_instance = AES256_CTR_DRBG(seed=seed)
            keygen_rand = drbg_instance.random_bytes(3*alg.n)
            sk_seed = keygen_rand[:alg.n]
            sk_prf = keygen_rand[alg.n:2*alg.n]
            pk_seed = keygen_rand[2*alg.n:]
            pk, sk = alg.slh_keygen_internal(sk_seed, sk_prf, pk_seed)
            signature_rand = sign(alg.slh_sign_internal, sk, msg, drbg_instance.random_bytes, alg.n, deterministic=False)
            signature_det = sign(alg.slh_sign_internal, sk, msg, drbg_instance.random_bytes, alg.n, deterministic=True)

            print("SphincsParameterSet = %s" % param, file=output)
            print("seed = %s" % seed.hex().upper(), file=output)
            print("msg = %s" % msg.hex().upper(), file=output)
            print("pk = %s" % pk.hex().upper(), file=output)
            print("sk = %s" % sk.hex().upper(), file=output)
            if param == "SLH-DSA-SHAKE-128s":
                print("SigDet = %s" % binascii.hexlify(signature_det).decode("utf-8").upper(), file=output)
            print("HashSigDet = %s" % binascii.hexlify(hash_fn(signature_det)).decode("utf-8").upper(), file=output)
            print("HashSigRand = %s" % binascii.hexlify(hash_fn(signature_rand)).decode("utf-8").upper(), file=output)
            print("", file=output)

    elif algo == "ml_dsa":
        for param in ML_DSA_PARAM.keys():
            for det_str, deterministic in [("Deterministic", True), ("Randomized", False)]:
                alg = ML_DSA(param)
                output = open('src/tests/data/pubkey/%s_%s.vec' % (transform_mldsa_string(param), det_str), 'w')
                print("# See src/scripts/dev_tools/gen_pqc_dsa_kats.py", file=output)
                print("", file=output)
                print(create_mldsa_ini_label(param, det_str), file=output)
                print("", file=output)

                hash_fn = sha3_256
                def mldsa_sign_internal(m, sk, rnd):
                    # For some reason the interfaces vary between FIPS 204 and FIPS 205...
                    if rnd is None:
                        rnd = bytes([0]*32)
                    return alg.sign_internal(sk, m, rnd)

                samples = 25
                for _ in range(100):
                    seed = drbg.random_bytes(len(entropy))

                    if samples == 0:
                        continue

                    samples -= 1
                    drbg_instance = AES256_CTR_DRBG(seed=seed)
                    keygen_rand = drbg_instance.random_bytes(32)
                    pk, sk = alg.keygen_internal(keygen_rand)
                    signature = sign(mldsa_sign_internal, sk, msg, drbg_instance.random_bytes, 32, deterministic=deterministic)

                    print("Seed = %s" % seed.hex().upper(), file=output)
                    print("Msg = %s" % msg.hex().upper(), file=output)
                    print("HashPk = %s" % binascii.hexlify(hash_fn(pk)).decode("utf-8").upper(), file=output)
                    print("HashSk = %s" % binascii.hexlify(hash_fn(keygen_rand)).decode("utf-8").upper(), file=output)
                    print("HashSig = %s" % binascii.hexlify(hash_fn(signature)).decode("utf-8").upper(), file=output)
                    print("", file=output)
    else:
        print("Usage: gen_pqc_dsa_kats.py <ml_dsa|slh_dsa>")
        return 1

if __name__ == '__main__':
    sys.exit(main())




#!/usr/bin/env python3

"""
Implements the "make check" target

(C) 2020 Jack Lloyd, Rene Meusel

Botan is released under the Simplified BSD License (see license.txt)
"""

import json
import logging
import optparse # pylint: disable=deprecated-module
import os
import subprocess
import sys

def run_and_check(cmd_line, env=None, cwd=None):

    logging.info("Starting %s", ' '.join(cmd_line))

    try:
        proc = subprocess.Popen(cmd_line, cwd=cwd, env=env)
        proc.communicate()
    except OSError as e:
        logging.error("Executing %s failed (%s)", ' '.join(cmd_line), e)

    if proc.returncode != 0:
        logging.error("Error running %s", ' '.join(cmd_line))
        sys.exit(1)


def make_environment(build_shared_lib):
    if not build_shared_lib:
        return None

    env = os.environ.copy()

    def extend_colon_list(k, n):
        env[k] = n if k not in env else ":".join([env[k], n])

    extend_colon_list("DYLD_LIBRARY_PATH", os.path.abspath("."))
    extend_colon_list("LD_LIBRARY_PATH", os.path.abspath("."))

    return env


def parse_options(args):
    parser = optparse.OptionParser()
    parser.add_option('--build-dir', default='build', metavar='DIR',
                      help='specify the botan build directory (default %default)')

    (options, args) = parser.parse_args(args)

    if len(args) > 1:
        raise Exception("Unknown arguments")

    return options


def read_config(config):
    try:
        with open(config, encoding='utf8') as f:
            return json.load(f)
    except OSError as ex:
        raise Exception('Failed to load build config %s - is build dir correct?' % (config)) from ex


def main(args=None):
    if args is None:
        args = sys.argv

    options = parse_options(args)

    cfg = read_config(os.path.join(options.build_dir, 'build_config.json'))

    test_exe = cfg.get('test_exe')
    build_shared_lib = cfg.get('build_shared_lib')

    if not os.path.isfile(test_exe) or not os.access(test_exe, os.X_OK):
        raise Exception("Test binary not built")

    run_and_check([test_exe, "--data-dir=%s" % cfg.get('test_data_dir')],
                  make_environment(build_shared_lib))

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
Compare Botan with OpenSSL using their respective benchmark utils

(C) 2017,2022 Jack Lloyd
    2023 René Meusel, Rohde & Schwarz Cybersecurity GmbH
    2023,2024 René Fischer

Botan is released under the Simplified BSD License (see license.txt)
"""

import logging
import os
import sys
import optparse # pylint: disable=deprecated-module
import subprocess
import re
import json
import math
import platform
from datetime import datetime
import shutil
import numpy as np
import matplotlib.pyplot as plt

def setup_logging(options):
    if options.verbose:
        log_level = logging.DEBUG
    elif options.quiet:
        log_level = logging.WARNING
    else:
        log_level = logging.INFO

    class LogOnErrorHandler(logging.StreamHandler):
        def emit(self, record):
            super().emit(record)
            if record.levelno >= logging.ERROR:
                sys.exit(1)

    lh = LogOnErrorHandler(sys.stdout)
    lh.setFormatter(logging.Formatter('%(levelname) 7s: %(message)s'))
    logging.getLogger().addHandler(lh)
    logging.getLogger().setLevel(log_level)

def run_command(cmd):
    logging.debug("Running '%s'", ' '.join(cmd))

    proc = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()

    if proc.returncode != 0:
        logging.error("Running command %s failed ret %d", ' '.join(cmd), proc.returncode)

    return stdout + stderr

def get_openssl_version(openssl):
    output = run_command([openssl, 'version'])

    openssl_version_re = re.compile(r'OpenSSL ([0-9a-z\.]+) .*')

    match = openssl_version_re.match(output)

    if match:
        return match.group(1)
    else:
        logging.warning("Unable to parse OpenSSL version output %s", output)
        return output

def get_botan_version(botan):
    return run_command([botan, 'version']).strip()

EVP_MAP = {
    'AES-128/GCM': 'aes-128-gcm',
    'AES-256/GCM': 'aes-256-gcm',
    'ChaCha20': 'chacha20',
    'SHA-1': 'sha1',
    'SHA-256': 'sha256',
    'SHA-384': 'sha384',
    'SHA-512': 'sha512',
    'SHA-3(256)': 'sha3-256',
    'SHA-3(512)': 'sha3-512'
    }

SIGNATURE_EVP_MAP = {
    'RSA': 'rsa',
    'ECDSA': 'ecdsa'
    }

KEY_AGREEMENT_EVP_MAP = {
    'DH': 'ffdh',
    'ECDH': 'ecdh'
    }

def run_openssl_bench(openssl, algo):

    logging.info('Running OpenSSL benchmark for %s', algo)

    cmd = [openssl, 'speed', '-seconds', '1', '-mr']

    if algo in EVP_MAP:
        cmd += ['-evp', EVP_MAP[algo]]
    elif algo in SIGNATURE_EVP_MAP:
        cmd += [SIGNATURE_EVP_MAP[algo]]
    elif algo in KEY_AGREEMENT_EVP_MAP:
        cmd += [KEY_AGREEMENT_EVP_MAP[algo]]
    else:
        cmd += [algo]

    output = run_command(cmd)
    results = []

    if algo in EVP_MAP:
        buf_header = re.compile(r'\+DT:([a-zA-Z0-9-]+):([0-9]+):([0-9]+)$')
        res_header = re.compile(r'\+R:([0-9]+):[a-zA-Z0-9-]+:([0-9]+\.[0-9]+)$')
        ignored = re.compile(r'\+(H|F):.*')

        result = {}

        for line in output.splitlines():
            if ignored.match(line):
                continue

            if not result:
                match = buf_header.match(line)
                if match is None:
                    logging.error("Unexpected output from OpenSSL %s", line)

                result = {'algo': algo, 'buf_size': int(match.group(3))}
            else:
                match = res_header.match(line)

                result['bytes'] = int(match.group(1)) * result['buf_size']
                result['runtime'] = float(match.group(2))
                result['bps'] = int(result['bytes'] / result['runtime'])
                results.append(result)
                result = {}
    elif algo in SIGNATURE_EVP_MAP:
        signature_ops = re.compile(r'\+(R1|R7|R5):([0-9]+):([0-9]+):([0-9]+\.[0-9]+)$')
        verify_ops    = re.compile(r'\+(R2|R8|R6):([0-9]+):([0-9]+):([0-9]+\.[0-9]+)$')
        ignored = re.compile(r'\+(DTP|F2|R3|R4|F4):.*')

        result = {}

        for line in output.splitlines():
            if ignored.match(line):
                continue

            if match := signature_ops.match(line):
                results.append({
                    'algo': algo,
                    'key_size': int(match.group(3)),
                    'op': 'sign',
                    'ops': int(match.group(2)),
                    'runtime': float(match.group(4))})
            elif match := verify_ops.match(line):
                results.append({
                    'algo': algo,
                    'key_size': int(match.group(3)),
                    'op': 'verify',
                    'ops': int(match.group(2)),
                    'runtime': float(match.group(4))
                })
            else:
                logging.error("Unexpected output from OpenSSL %s", line)

    elif algo in KEY_AGREEMENT_EVP_MAP:
        res_header    = re.compile(r'\+(R7|R9|R12|R14):([0-9]+):([0-9]+):([0-9]+\.[0-9]+)$')
        ignored = re.compile(r'\+(DTP|F5|F8):.*')

        result = {}

        for line in output.splitlines():
            if ignored.match(line):
                continue

            if match := res_header.match(line):
                results.append({
                    'algo': algo,
                    'key_size': int(match.group(3)),
                    'ops': int(match.group(2)),
                    'runtime': float(match.group(4))})
            else:
                logging.error("Unexpected output from OpenSSL %s", line)

    return results

def run_botan_bench(botan, runtime, buf_sizes, algo):

    logging.info('Running Botan benchmark for %s', algo)

    cmd = [botan, 'speed', '--format=json', '--msec=%d' % int(runtime * 1000),
           '--buf-size=%s' % (','.join([str(i) for i in buf_sizes])), algo]
    output = run_command(cmd)
    output = json.loads(output)

    return output

def run_botan_signature_bench(botan, runtime, algo):

    logging.info('Running Botan benchmark for %s', algo)

    cmd = [botan, 'speed', '--format=json', '--msec=%d' % int(runtime * 1000), algo]
    output = run_command(cmd)
    output = json.loads(output)

    results = []
    for verify in output:
        for sign in output:
            if sign['op'] == 'sign' and verify['op'] == 'verify' and verify['algo'] == sign['algo']:
                results.append({
                    'algo': algo,
                    'key_size': int(re.search(r'[A-Z]+-[a-z]*([0-9]+).*', sign['algo']).group(1)),
                    'op': 'sign',
                    'ops': sign['events'],
                    'runtime': sign['nanos'] / 1000 / 1000 / 1000,
                })
                results.append({
                    'algo': algo,
                    'key_size': int(re.search(r'[A-Z]+-[a-z]*([0-9]+).*', sign['algo']).group(1)),
                    'op': 'verify',
                    'ops': verify['events'],
                    'runtime': verify['nanos'] / 1000 / 1000 / 1000,
                })
    return results

def run_botan_key_agreement_bench(botan, runtime, algo):

    logging.info('Running Botan benchmark for %s', algo)

    cmd = [botan, 'speed', '--format=json', '--msec=%d' % int(runtime * 1000), algo]
    output = run_command(cmd)
    output = json.loads(output)

    results = []
    for res in output:
        if res['op'] == 'key agreements':
            results.append({
                'algo': algo,
                'key_size': int(re.search(r'[A-Z]+-[a-z]*([0-9]+).*', res['algo']).group(1)),
                'ops': res['events'],
                'runtime': res['nanos'] / 1000 / 1000 / 1000,
            })
    return results

class BenchmarkResult:
    def __init__(self, algo, sizes, openssl_results, botan_results):
        self.algo = algo
        self.results = {}

        def find_result(results, sz):
            for r in results:
                if 'buf_size' in r and r['buf_size'] == sz:
                    return r['bps']
            raise Exception("Could not find expected result in data")

        for size in sizes:
            self.results[size] = {
                'openssl': find_result(openssl_results, size),
                'botan': find_result(botan_results, size)
            }

    def result_string(self):

        out = ""
        for (k, v) in self.results.items():

            if v['openssl'] > v['botan']:
                winner = 'openssl'
                ratio = float(v['openssl']) / v['botan']
            else:
                winner = 'botan'
                ratio = float(v['botan']) / v['openssl']

            out += "algo %s buf_size % 6d botan % 12d bps openssl % 12d bps adv %s by %.02f\n" % (
                self.algo, k, v['botan'], v['openssl'], winner, ratio)
        return out

    def graphs(self):
        buf_sizes = self.results.keys()
        x = np.arange(len(buf_sizes))
        width = 0.4
        multiplier = 0

        fig, ax = plt.subplots(layout='constrained')

        for lib in ['Botan', 'OpenSSL']:
            offset = width * multiplier
            results = [math.ceil(x[lib.lower()]/(1024*1024)) for x in self.results.values()]
            rects = ax.bar(x + offset, results, width, label=lib)
            ax.bar_label(rects, padding=5)
            multiplier += 1

        ax.set_xlabel('Buffer size (bytes)')
        ax.set_ylabel('MiB/sec')
        ax.set_title(self.algo)
        ax.set_xticks(x + width, buf_sizes)
        ax.legend(loc='upper left', ncols=2)

        return {self.algo: fig}


class SignatureBenchmarkResult:
    def __init__(self, algo, sizes, openssl_results, botan_results):
        self.algo = algo
        self.results = {}

        def find_result(results, sz):
            for verify in results:
                for sign in results:
                    if sign['op'] == 'sign' and verify['op'] == 'verify' and verify['key_size'] == sz:
                        return {'sign': sign['ops'], 'verify': verify['ops']}
            raise Exception("Could not find expected result in data")

        for size in sizes:
            self.results[size] = {
                'openssl': find_result(openssl_results, size),
                'botan': find_result(botan_results, size)
            }

    def result_string(self):

        out = ""
        for (k, v) in self.results.items():
            openssl = v['openssl']
            botan = v['botan']

            for op in openssl.keys():
                if openssl[op] > botan[op]:
                    winner = 'openssl'
                    ratio = float(openssl[op]) / botan[op]
                else:
                    winner = 'botan'
                    ratio = float(botan[op]) / openssl[op]

                out += "algo %s key_size % 5d % 8s botan % 10d openssl % 10d adv %s by %.02f\n" % (
                        self.algo, k, op, botan[op], openssl[op], winner, ratio)

        return out

    def graphs(self):
        key_sizes = self.results.keys()
        x = np.arange(len(key_sizes))
        width = 0.4

        graphs = {}

        for op in ['sign', 'verify']:
            fig, ax = plt.subplots(layout='constrained')
            multiplier = 0

            for lib in ['Botan', 'OpenSSL']:
                offset = width * multiplier
                results = [x[lib.lower()][op] for x in self.results.values()]
                rects = ax.bar(x + offset, results, width, label=lib)
                ax.bar_label(rects, padding=5)
                multiplier += 1

            # Add some text for labels, title and custom x-axis tick labels, etc.
            ax.set_xlabel('Key size (bits)')
            ax.set_ylabel('Ops/sec')
            ax.set_title(self.algo + ' ' + op)
            ax.set_xticks(x + width, key_sizes)
            ax.legend(loc='upper right', ncols=2)

            graphs.update({self.algo + '_' + op: fig})

        return graphs


class KeyAgreementBenchmarkResult:
    def __init__(self, algo, sizes, openssl_results, botan_results):
        self.algo = algo
        self.results = {}

        def find_result(results, sz):
            for r in results:
                if 'key_size' in r and r['key_size'] == sz:
                    return r['ops']
            raise Exception("Could not find expected result in data")

        for size in sizes:
            self.results[size] = {
                'openssl': find_result(openssl_results, size),
                'botan': find_result(botan_results, size)
            }

    def result_string(self):

        out = ""
        for (k, v) in self.results.items():

            if v['openssl'] > v['botan']:
                winner = 'openssl'
                ratio = float(v['openssl']) / v['botan']
            else:
                winner = 'botan'
                ratio = float(v['botan']) / v['openssl']

            out += "algo %s key_size % 6d botan % 12d key agreements openssl % 12d key agreements adv %s by %.02f\n" % (
                self.algo, k, v['botan'], v['openssl'], winner, ratio)
        return out

    def graphs(self):
        key_sizes = self.results.keys()
        x = np.arange(len(key_sizes))
        width = 0.4
        multiplier = 0

        fig, ax = plt.subplots(layout='constrained')

        for lib in ['Botan', 'OpenSSL']:
            offset = width * multiplier
            results = [x[lib.lower()] for x in self.results.values()]
            rects = ax.bar(x + offset, results, width, label=lib)
            ax.bar_label(rects, padding=5)
            multiplier += 1

        ax.set_xlabel('Key size (bits)')
        ax.set_ylabel('Key agreements/sec')
        ax.set_title(self.algo)
        ax.set_xticks(x + width, key_sizes)
        ax.legend(loc='upper right', ncols=2)

        return {self.algo: fig}


def bench_algo(openssl, botan, algo):
    openssl_results = run_openssl_bench(openssl, algo)

    buf_sizes = sorted([x['buf_size'] for x in openssl_results])
    runtime = sum(x['runtime'] for x in openssl_results) / len(openssl_results) / len(buf_sizes)

    botan_results = run_botan_bench(botan, runtime, buf_sizes, algo)

    return BenchmarkResult(algo, buf_sizes, openssl_results, botan_results)

def bench_signature_algo(openssl, botan, algo):
    openssl_results = run_openssl_bench(openssl, algo)

    runtime = sum(x['runtime'] for x in openssl_results) / len(openssl_results)
    botan_results = run_botan_signature_bench(botan, runtime, algo)

    kszs_ossl = {x['key_size'] for x in openssl_results}
    kszs_botan = {x['key_size'] for x in botan_results}

    return SignatureBenchmarkResult(algo, sorted(kszs_ossl.intersection(kszs_botan)), openssl_results, botan_results)

def bench_key_agreement_algo(openssl, botan, algo):
    openssl_results = run_openssl_bench(openssl, algo)

    runtime = sum(x['runtime'] for x in openssl_results) / len(openssl_results)
    botan_results = run_botan_key_agreement_bench(botan, runtime, algo)

    kszs_ossl = {x['key_size'] for x in openssl_results}
    kszs_botan = {x['key_size'] for x in botan_results}

    return KeyAgreementBenchmarkResult(algo, sorted(kszs_ossl.intersection(kszs_botan)), openssl_results, botan_results)


def gen_html_report(botan_version, openssl_version, cmdline):
    html = """<!DOCTYPE html>
    <html>
    <head>
    <title>Botan Benchmark HTML Report</title>
    </head>
    <body>
    <h1>Botan Benchmark HTML Report</h1>
    <h2>Environment</h2>
    <p><b>Date: </b> {date}</p>
    <p><b>Operating System:</b> {os_info}</p>
    <p><b>Botan version:</b> {botan_version}</p>
    <details>
        <summary><b>Build Info</b></summary>
        <p><pre>{build_h}</pre></p>
    </details>
    <p><b>OpenSSL version:</b> {openssl_version}</p>
    <p><b>Invocation:</b>
    <pre>{cmdline}</pre>
    </p>

    <h2>Results</h2>"""

    for img in sorted(os.listdir('bench_html/img')):
        html += '<img src="img/%s">' % img

    html += """
    </body>
    </html>
    """

    os_info = platform.system()
    os_info += ' ' + platform.release()
    os_info += ' ('
    os_info += platform.machine()
    os_info += ')'

    include_path = re.search(r'-I(.*)', run_command(['botan', 'config', 'cflags']).strip()).group(1)
    build_h = 'N/A'

    try:
        with open(os.path.join(include_path, 'botan/build.h'), 'r', encoding='utf-8') as f:
            build_h = f.read()
    except Exception as e:
        logging.warning("Unable to read build.h: %s", e)

    with open('bench_html/index.html', 'w', encoding='utf-8') as f:
        f.write(html.format(date=datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3],
                            os_info=os_info,
                            botan_version=botan_version,
                            build_h=build_h,
                            openssl_version=openssl_version,
                            cmdline=cmdline))


def main(args=None):
    if args is None:
        args = sys.argv

    usage = "usage: %prog [options] [algo1] [algo2] [...]"
    parser = optparse.OptionParser(usage=usage)

    parser.add_option('--verbose', action='store_true', default=False, help="be noisy")
    parser.add_option('--quiet', action='store_true', default=False, help="be very quiet")

    parser.add_option('--html-report', action='store_true', default=False, help="create HTML report")

    parser.add_option('--openssl-cli', metavar='PATH',
                      default='/usr/bin/openssl',
                      help='Path to openssl binary (default %default)')

    parser.add_option('--botan-cli', metavar='PATH',
                      default='/usr/bin/botan',
                      help='Path to botan binary (default %default)')

    (options, args) = parser.parse_args(args)

    setup_logging(options)

    openssl = options.openssl_cli
    botan = options.botan_cli

    if os.access(openssl, os.X_OK) is False:
        logging.error("Unable to access openssl binary at %s", openssl)

    if os.access(botan, os.X_OK) is False:
        logging.error("Unable to access botan binary at %s", botan)

    openssl_version = get_openssl_version(openssl)
    botan_version = get_botan_version(botan)

    logging.info("Comparing Botan %s with OpenSSL %s", botan_version, openssl_version)

    if options.html_report:
        if os.path.exists('bench_html'):
            shutil.rmtree('bench_html')

        os.mkdir('bench_html')
        os.mkdir('bench_html/img')

    graphs = {}

    if len(args) > 1:
        algos = args[1:]
        for algo in algos:
            if algo in EVP_MAP:
                result = bench_algo(openssl, botan, algo)
                graphs.update(result.graphs())
                print(result.result_string())
            elif algo in SIGNATURE_EVP_MAP:
                result = bench_signature_algo(openssl, botan, algo)
                graphs.update(result.graphs())
                print(result.result_string())
            elif algo in KEY_AGREEMENT_EVP_MAP:
                result = bench_key_agreement_algo(openssl, botan, algo)
                graphs.update(result.graphs())
                print(result.result_string())
            else:
                logging.error("Unknown algorithm '%s'", algo)
    else:
        for algo in sorted(EVP_MAP):
            result = bench_algo(openssl, botan, algo)
            graphs.update(result.graphs())
            print(result.result_string())

        for algo in sorted(SIGNATURE_EVP_MAP):
            result = bench_signature_algo(openssl, botan, algo)
            graphs.update(result.graphs())
            print(result.result_string())

        for algo in sorted(KEY_AGREEMENT_EVP_MAP):
            result = bench_key_agreement_algo(openssl, botan, algo)
            graphs.update(result.graphs())
            print(result.result_string())

    if options.html_report:
        for title, graph in graphs.items():
            graph.savefig('bench_html/img/' + title.replace('/', '--') + ".png")

        gen_html_report(botan_version, openssl_version, ' '.join(sys.argv))

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
A script to automatically write docs to /docs. Currently it generates
os.rst, a feature table of OS features.

Requires Python 3.

(C) 2015 Simon Warta (Kullo GmbH)

Botan is released under the Simplified BSD License (see license.txt)
"""

# global
import argparse
import datetime
import glob
import os
import sys

# Assume this script is in botan/src/scripts/dev_tools
botan_root = os.path.join(os.path.dirname(sys.argv[0]), "..", "..", "..")

# locale
sys.path.append(botan_root)
from configure import OsInfo # noqa: E402

parser = argparse.ArgumentParser(description="")
parser.add_argument('--verbose', dest='verbose', action='store_const',
                    const=True, default=False,
                    help='Verbose output (default: false)')
args = parser.parse_args()

def update_os():
    TABLE_TITLE="OS Features"

    files = []
    files += glob.glob(botan_root + '/src/build-data/os/*.txt')
    files.sort()

    if len(files) == 0:
        print("No info.txt files found.")
        sys.exit(1)

    f1 = open(os.path.join(botan_root, 'doc', 'dev_ref', 'os.rst'), 'w+')

    all_features = set()
    oss = {}

    for filename in files:
        o = OsInfo(filename)
        oss[o.basename] = o
        all_features |= set(o.target_features)
        if args.verbose:
            print(o.basename)
            print(o.target_features)

    featurelist = list(all_features)
    featurelist.sort()
    oslist = list(oss.keys())
    oslist.sort()

    if args.verbose:
        print(featurelist)

    today = datetime.date.today().strftime("%Y-%m-%d")
    print(".. This file was automatically generated by src/scripts/dev_tools/%s on %s"
                    % (os.path.basename(sys.argv[0]), today), file=f1)
    print(".. All manual changes will be lost. Edit the script instead.", file=f1)
    print("", file=f1)
    print(TABLE_TITLE, file=f1)
    print("========================================", file=f1)
    print("", file=f1)

    print("A summary of OS features as defined in ``src/build-data/os``.", file=f1)
    print("", file=f1)

    print("::", file=f1)
    print("", file=f1)
    for o in oslist:
        print("  %s: %s" % (o[0:1], o), file=f1)
    print("", file=f1)

    print('.. csv-table::', file=f1)
    print('   :header: "Feature", "' + '", "'.join([o[0:1] for o in oslist]) + '"', file=f1)
    print('', file=f1)

    for f in featurelist:
        line = '   "' + f + '"'
        for o in oslist:
            line += ', "'
            line += 'X' if f in oss[o].target_features else ' '
            line += '"'
        print(line, file=f1)
if __name__ == '__main__':
    update_os()



#!/usr/bin/env python3

"""
This configures and builds with many different sub-configurations
in an attempt to flush out missing feature macro checks, etc.

There is probably no reason for you to run this. Unless you want to.

(C) 2017 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import optparse # pylint: disable=deprecated-module
import sys
import subprocess

def get_module_list(configure_py):
    configure = subprocess.Popen([configure_py, '--list-modules'], stdout=subprocess.PIPE)

    (stdout, _) = configure.communicate()

    if configure.returncode != 0:
        raise Exception("Running configure.py --list-modules failed")

    modules = [s.decode('ascii') for s in stdout.split()]
    modules.remove('tpm') # can't test
    modules.remove('jitter_rng')
    modules.remove('esdm_rng')
    modules.remove('tpm2')
    modules.remove('tpm2_crypto_backend')
    modules.remove('tpm2_ecc')
    modules.remove('tpm2_rsa')
    modules.remove('base') # can't remove
    return modules

def get_concurrency():
    def_concurrency = 2

    try:
        import multiprocessing
        return max(def_concurrency, multiprocessing.cpu_count())
    except ImportError:
        return def_concurrency

def try_to_run(cmdline):
    print("Running %s ... " % (' '.join(cmdline)))
    sys.stdout.flush()

    cmd = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = cmd.communicate()

    failed = (cmd.returncode != 0)

    if failed:
        print("FAILURE")
        print(stdout)
        print(stderr)
        sys.stdout.flush()

    return not failed

def run_test_build(configure_py, modules, include, jobs, run_tests):
    config = [configure_py, '--without-documentation', '--compiler-cache=ccache', '--maintainer-mode']

    if include:
        config.append('--minimized')
        if modules:
            config.append('--enable-modules=' + ','.join(modules))
    else:
        config.append('--disable-modules=' + ','.join(modules))

    if try_to_run(config) is False:
        return False

    if try_to_run(['make', '-j', str(jobs)]) is False:
        return False

    if run_tests is False:
        return True

    # Flaky test causing errors when running tests
    tests_to_skip = []

    cmdline = ['./botan-test', '--test-threads=%d' % (jobs)]

    if len(tests_to_skip) > 0:
        cmdline.append('--skip-tests=%s' % (','.join(tests_to_skip)))

    return try_to_run(cmdline)

def main(args):

    # TODO take configure.py and botan-test paths via options

    parser = optparse.OptionParser()

    parser.add_option('--run-tests', default=False, action='store_true')
    parser.add_option('--jobs', default=get_concurrency(),
                      help="jobs to run (default %default)")

    (options, args) = parser.parse_args(args)

    run_tests = options.run_tests
    jobs = int(options.jobs)

    configure_py = './configure.py'
    modules = get_module_list(configure_py)

    cant_disable = ['block', 'hash', 'hex', 'mac', 'modes', 'rng', 'stream', 'utils', 'cpuid']
    always_include = ['thread_utils', 'sha2_64']

    fails = 0
    failed = []

    for module in sorted(modules):
        if (module in always_include) or (module in cant_disable):
            continue # already testing it

        extra = []
        if module == 'auto_rng':
            extra.append('system_rng')
        if run_test_build(configure_py, [module] + always_include + extra, True, jobs, run_tests) is False:
            failed.append(module)
            fails += 1

    for module in sorted(modules):
        if module in cant_disable or module in always_include:
            continue
        if run_test_build(configure_py, [module], False, jobs, run_tests) is False:
            failed.append(module)
            fails += 1

    if len(failed) > 0:
        print("Failed building with %s" % (' '.join(failed)))
    else:
        print("All configurations ok")

    return fails

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python3

#
# Strips the KAT harness in the FrodoKEM reference implementation down
# to a less space consuming version. This script was used to generate
# `src/tests/data/pubkey/frodokem_kat.vec` test data from the *.rsp files in
# the reference implementation repository.
#
# See here: https://github.com/microsoft/PQCrypto-LWEKE/tree/master
#
# (C) 2023 Jack Lloyd
# (C) 2023 René Meusel, Amos Treiber - Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import sys
import hashlib
import binascii
import os

class KatReader:
    def __init__(self, file):
        self.file = file
        self.last_mlen = None

    def next_value(self):
        while True:
            line = self.file.readline()

            if line == "":
                return (None, None)

            if line.startswith('#') or line == "\n":
                continue

            key, val = line.strip().split(' = ')

            return (key, val)

    def read_kats(self):
        kat = {}

        while True:
            key, val = self.next_value()

            if key is None:
                return # eof

            if key not in ['count', 'seed', 'pk', 'sk', 'ct', 'ss']:
                raise Exception("Unknown key %s" % (key))

            if key == 'count':
                kat[key] = int(val)
            else:
                kat[key] = val

            if key == 'ss':
                yield kat
                kat = {}

def shake_256_16(v):
    # v is assumed to be hex
    h = hashlib.shake_256()
    h.update(binascii.unhexlify(v))
    return h.hexdigest(16)

def compress_kat(kat):
    del kat['count'] # not needed

    # rename keys
    kat['Seed'] = kat.pop('seed')
    kat['SS']   = kat.pop('ss')
    kat['PK']   = shake_256_16(kat.pop('pk'))
    kat['SK']   = shake_256_16(kat.pop('sk'))
    kat['CT']   = shake_256_16(kat.pop('ct'))

    return kat

def map_mode(mode, is_ephemeral = False):
    out = None
    if mode == "PQCkemKAT_19888":
        out = "eFrodoKEM-640-AES"
    if mode == "PQCkemKAT_19888_shake":
        out = "eFrodoKEM-640-SHAKE"
    if mode == "PQCkemKAT_31296":
        out = "eFrodoKEM-976-AES"
    if mode == "PQCkemKAT_31296_shake":
        out = "eFrodoKEM-976-SHAKE"
    if mode == "PQCkemKAT_43088":
        out = "eFrodoKEM-1344-AES"
    if mode == "PQCkemKAT_43088_shake":
        out = "eFrodoKEM-1344-SHAKE"

    if out is None:
        raise Exception('Unknown FrodoKEM mode', mode)

    if is_ephemeral:
        return out
    else:
        return out[1:]  # remove 'e' to obtain 'FrodoKEM'

def main(args = None):
    """Set True for eFrodo, otherwise the non-ephemeral variant is assumed.
    Necessary because the files have the same name for either variant."""
    is_ephemeral = False

    if args is None:
        args = sys.argv

    with open('src/tests/data/pubkey/frodokem_kat.vec', 'w') as output:
        print("# This file was auto-generated from the reference implementation's KATs", file=output)
        print("# See src/scripts/dev_tools/gen_frodo_kat.py\n", file=output)

        for file in args[1:]:
            mode = map_mode(os.path.basename(os.path.splitext(file)[0]), is_ephemeral)

            reader = KatReader(open(file))

            print(f"[{mode}]", file=output)

            for kat in list(reader.read_kats())[:25]:
                kat = compress_kat(kat)

                for key in kat.keys():
                    print(key, '=', kat[key], file=output)
                print("", file=output)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2015,2016,2017,2018 Simon Warta
(C) 2019,2020,2021 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import argparse
import re
import subprocess
import sys
import os.path
import logging
import time
from collections import OrderedDict
import multiprocessing
from multiprocessing.pool import ThreadPool

class VecDocument:
    def __init__(self, filepath):
        self.data = OrderedDict()
        last_testcase_number = 1
        current_testcase_number = 1
        current_group_name = ""
        last_group_name = ""
        current_testcase = {}

        PATTERN_GROUPHEADER = r"^\[(.+)\]$"
        PATTERN_KEYVALUE = r"^\s*([a-zA-Z]+)\s*=(.*)$"

        with open(filepath, 'r') as f:
            # Append one empty line to simplify parsing
            lines = f.read().splitlines() + ["\n"]

            for line in lines:
                line = line.strip()
                if line.startswith("#"):
                    pass # Skip
                elif line == "":
                    current_testcase_number += 1
                elif re.match(PATTERN_GROUPHEADER, line):
                    match = re.match(PATTERN_GROUPHEADER, line)
                    current_group_name = match.group(1)
                elif re.match(PATTERN_KEYVALUE, line):
                    match = re.match(PATTERN_KEYVALUE, line)
                    key = match.group(1)
                    value = match.group(2).strip()
                    current_testcase[key] = value

                if current_testcase_number != last_testcase_number:
                    if current_group_name not in self.data:
                        self.data[current_group_name] = []
                    if len(current_testcase) != 0:
                        self.data[current_group_name].append(current_testcase)
                    current_testcase = {}
                    last_testcase_number = current_testcase_number

                if current_group_name != last_group_name:
                    last_group_name = current_group_name
                    # Reset testcase number
                    last_testcase_number = 1
                    current_testcase_number = 1

    def get_data(self):
        return self.data

TESTS_RUN = 0
TESTS_FAILED = 0

class TestLogHandler(logging.StreamHandler, object):
    def emit(self, record):
        # Do the default stuff first
        super(TestLogHandler, self).emit(record)
        if record.levelno >= logging.ERROR:
            global TESTS_FAILED
            TESTS_FAILED += 1

def setup_logging(options):
    if options.verbose:
        log_level = logging.DEBUG
    elif options.quiet:
        log_level = logging.WARNING
    else:
        log_level = logging.INFO

    lh = TestLogHandler(sys.stdout)
    lh.setFormatter(logging.Formatter('%(levelname) 7s: %(message)s'))
    logging.getLogger().addHandler(lh)
    logging.getLogger().setLevel(log_level)

def test_cipher_kat(cli_binary, data):
    iv = data['Nonce'] if 'Nonce' in data else ''
    key = data['Key']
    plaintext = data['In'].lower()
    ciphertext = data['Out'].lower()
    algorithm = data['Algorithm']
    direction = data['Direction']

    cmd = [
        cli_binary,
        "cipher",
        "--cipher=%s" % algorithm,
        "--nonce=%s" % iv,
        "--key=%s" % key,
        "-"]

    if 'AD' in data:
        cmd += ['--ad=%s' % (data['AD'])]

    if direction == "decrypt":
        cmd += ['--decrypt']

    if direction == "decrypt":
        invalue = ciphertext
    else:
        invalue = plaintext

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout_raw, stderr_raw) = p.communicate(input=binascii.unhexlify(invalue))
    output = binascii.hexlify(stdout_raw).decode("UTF-8").lower()
    stderr = stderr_raw.decode("UTF-8")

    if stderr != '':
        logging.error("Unexpected stderr output %s" % (stderr))

    expected = plaintext if direction == "decrypt" else ciphertext
    if expected != output:
        logging.error("For test %s got %s expected %s" % (data['testname'], output, expected))

def get_testdata(document, max_tests):
    out = []
    for algorithm in document:
        testcase_number = 0
        for testcase in document[algorithm]:
            testcase_number += 1
            for direction in ['encrypt', 'decrypt']:
                testname = "{} no {:0>3} ({})".format(
                    algorithm.lower(), testcase_number, direction)
                testname = re.sub("[^a-z0-9-]", "_", testname)
                testname = re.sub("_+", "_", testname)
                testname = testname.strip("_")
                test = {'testname': testname}
                for key in testcase:
                    value = testcase[key]
                    test[key] = value
                test['Algorithm'] = algorithm
                test['Direction'] = direction

                out.append(test)

            if max_tests > 0 and testcase_number > max_tests:
                break
    return out

def main(args=None):
    if args is None:
        args = sys.argv

    parser = argparse.ArgumentParser(description="")
    parser.add_argument('cli_binary', help='path to the botan cli binary')
    parser.add_argument('--threads', type=int, default=0, metavar="T")
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--quiet', action='store_true', default=False)
    parser.add_argument('--run-slow-tests', action='store_true', default=False)
    parser.add_argument('--test-data-dir', default='.')
    args = parser.parse_args()

    setup_logging(args)

    cli_binary = args.cli_binary
    max_tests = 0 if args.run_slow_tests else 30
    threads = args.threads

    if threads == 0:
        threads = multiprocessing.cpu_count()

    test_data_dir = os.path.join(args.test_data_dir, 'src', 'tests', 'data')

    mode_test_data = [
        os.path.join(test_data_dir, 'aead', 'ccm.vec'),
        os.path.join(test_data_dir, 'aead', 'chacha20poly1305.vec'),
        os.path.join(test_data_dir, 'aead', 'eax.vec'),
        os.path.join(test_data_dir, 'aead', 'gcm.vec'),
        os.path.join(test_data_dir, 'aead', 'ocb.vec'),
        os.path.join(test_data_dir, 'modes', 'cbc.vec'),
        os.path.join(test_data_dir, 'modes', 'cfb.vec'),
        os.path.join(test_data_dir, 'modes', 'ctr.vec'),
        os.path.join(test_data_dir, 'modes', 'xts.vec'),
    ]

    kats = []
    for f in mode_test_data:
        vecfile = VecDocument(f)
        kats += get_testdata(vecfile.get_data(), max_tests)

    start_time = time.time()

    if threads > 1:
        with ThreadPool(processes=threads) as pool:
            results = []
            for test in kats:
                results.append(pool.apply_async(test_cipher_kat, (cli_binary, test)))

            for result in results:
                result.get()
    else:
        for test in kats:
            test_cipher_kat(cli_binary, test)

    end_time = time.time()

    print("Ran %d tests with %d failures in %.02f seconds" % (
        len(kats), TESTS_FAILED, end_time - start_time))

    if TESTS_FAILED > 0:
        return 1
    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

import argparse
import os
from common import run_cmd, get_concurrency


BORING_REPO = "https://github.com/randombit/boringssl.git"
BORING_BRANCH = "rene/runner-20241016"

BORING_PATH = "build_deps/boringssl"
BOGO_PATH = os.path.join(BORING_PATH, "ssl", "test", "runner")

SHIM_PATH = "./botan_bogo_shim"
SHIM_CONFIG_NO_TLS13 = "src/bogo_shim/config_no_tls13.json"
SHIM_CONFIG_NO_TLS12 = "src/bogo_shim/config_no_tls12.json"
SHIM_CONFIG = "src/bogo_shim/config.json"


def main():
    parser = argparse.ArgumentParser(description='Run BoringSSL Bogo tests with Botan shim')
    parser.add_argument('--without-tls-12', action='store_true',
                        help='Use shim config that disables TLS 1.2')
    parser.add_argument('--without-tls-13', action='store_true',
                        help='Use shim config that disables TLS 1.3')
    parser.add_argument('--wait-for-debugger', action='store_true',
                        help='BoGo waits for some seconds so that we can attach a debugger to the shim')
    parser.add_argument('bogo_args', nargs=argparse.REMAINDER, help='Extra args for the bogo runner')
    args = parser.parse_args()

    # Select config depending on the option
    if args.without_tls_13:
        config_path = SHIM_CONFIG_NO_TLS13
    elif args.without_tls_12:
        config_path = SHIM_CONFIG_NO_TLS12
    else:
        config_path = SHIM_CONFIG

    if not os.path.isdir(BORING_PATH):
        # check out our fork of boring ssl
        run_cmd("git clone --depth 1 --branch %s %s %s" %
                (BORING_BRANCH, BORING_REPO, BORING_PATH))

    # make doubly sure we're on the correct branch
    run_cmd("git -C %s checkout %s" % (BORING_PATH, BORING_BRANCH))

    bogo_args = ';'.join(args.bogo_args) if args.bogo_args else ''
    extra_args = "-wait-for-debugger " if args.wait_for_debugger else ""
    extra_args += "-skip-tls12 -skip-dtls " if args.without_tls_12 else ""
    extra_args += "-skip-tls13 " if args.without_tls_13 else ""
    extra_args += "-debug -test '%s'" % bogo_args if bogo_args else ''

    run_cmd("go test -pipe -num-workers %d -shim-path %s -shim-config %s %s" %
            (get_concurrency(), os.path.abspath(SHIM_PATH), os.path.abspath(config_path), extra_args),
            BOGO_PATH)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# (C) 2025 Jack Lloyd
#     2025 René Meusel, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import subprocess
import os
import sys
import argparse
import tempfile
import platform

from textwrap import indent
from enum import Enum

SCRIPT_LOCATION = os.path.dirname(os.path.abspath(__file__))
GDB_EXTENSION = os.path.join(SCRIPT_LOCATION, "gdb", "strubtest.py")

class CPU(Enum):
    ANY = "any"
    X86_64 = "x86_64"
    AARCH64 = "aarch64"

    @staticmethod
    def current():
        cpu = platform.machine().lower()
        if cpu in ['x86_64', 'amd64']:
            return CPU.X86_64
        elif cpu in ['aarch64', 'arm64']:
            return CPU.AARCH64
        else:
            raise ValueError(f"Unsupported or unknown CPU architecture: {cpu}")

class StrubTest:
    def __init__(self, symbol, inferior_cmdline, masked_cpuid_bits = None, cpu = CPU.ANY, expect_fail = False):
        self.symbol = symbol
        self.inferior_cmdline = inferior_cmdline
        self.masked_cpuid_bits = masked_cpuid_bits
        self.cpu = cpu
        self.expect_fail = expect_fail

    @property
    def gdb_command(self):
        return ["gdb", "-x", GDB_EXTENSION,
                       "-ex", f"strubtest {self.symbol}",
                       "-ex", "run",
                       "--batch",
                       "--args", *self.inferior_cmdline]

    @property
    def rendered_gdb_command(self):
        return " ".join([token if " " not in token else f"'{token}'" for token in self.gdb_command])

    @property
    def environment(self):
        return {"BOTAN_CLEAR_CPUID": ",".join(self.masked_cpuid_bits)} if self.masked_cpuid_bits else {}

    @property
    def rendered_environment(self):
        full_env = os.environ.copy()
        full_env.update(self.environment)
        return full_env

    @property
    def runnable(self):
        return True if self.cpu == CPU.ANY else self.cpu == CPU.current()

    def run(self):
        print(f"Checking {self.symbol}... ", end="")
        if not self.runnable:
            return self._skip("incompatible platform")

        try:
            proc = subprocess.run(self.gdb_command, capture_output=True, check=False, env=self.rendered_environment)
        except subprocess.SubprocessError as ex:
            return self._fail("subprocess failure", exception=ex)

        if proc.returncode != 0:
            return self._fail("nonzero error code", proc_result=proc)

        if "Error: " in proc.stderr.decode("utf-8"):
            return self._fail("fail", proc_result=proc, may_be_expected=True)

        if "Success: " in proc.stdout.decode("utf-8"):
            return self._succeed(proc_result=proc)
        else:
            return self._fail("never invoked", proc_result=proc)

    def _print_debug_report(self, proc_result = None, exception = None):
        print(f"    ran: {self.rendered_gdb_command}")

        if self.environment:
            print("    Environment:")
            print(indent("\n".join([f"{k}={v}" for (k,v) in self.environment.items()]), " " * 6))
        if exception:
            print("    Exception:")
            print(indent(str(exception), " " * 6))
        if proc_result:
            if proc_result.stdout:
                print("    stdout:")
                print(indent(proc_result.stdout.decode("utf-8"), " " * 6))
            if proc_result.stderr:
                print("    stderr:")
                print(indent(proc_result.stderr.decode("utf-8"), " " * 6))

    def _fail(self, errmsg, proc_result = None, exception = None, may_be_expected = False):
        if may_be_expected and self.expect_fail:
            print(f"{errmsg} (expected)")
            return True
        else:
            print(errmsg)
            self._print_debug_report(proc_result=proc_result, exception=exception)
            return False

    def _succeed(self, proc_result = None):
        if not self.expect_fail:
            print("ok")
        else:
            print("ok (unexpected)")
            self._print_debug_report(proc_result=proc_result)
        return not self.expect_fail

    def _skip(self, reason):
        print(f"skipped ({reason})")
        return True

def dummy_file(size = 1024):
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(os.urandom(size))
        return temp_file.name

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--botan-cli', required=True)
    args = parser.parse_args()

    cli = args.botan_cli
    myfile = dummy_file()

    tests = [
        # This is a self-test, it is expected to fail because the version
        # information is naturally not annotated for stack scrubbing.
        StrubTest("Botan::version_string", [cli, "version", "--full"], expect_fail=True),

        # Below is a list of all strub-annotated symbols. Each of these
        # symbols is tested by running the Botan CLI with GDB and checking
        # that the stack scrubbing was performed correctly.

        StrubTest("Botan::SHA_256::compress_digest",          [cli, "hash", "--algo=SHA-256", myfile]),
        StrubTest("Botan::SHA_256::compress_digest_x86",      [cli, "hash", "--algo=SHA-256", myfile], cpu=CPU.X86_64),
        StrubTest("Botan::SHA_256::compress_digest_armv8",    [cli, "hash", "--algo=SHA-256", myfile], cpu=CPU.AARCH64),
        StrubTest("Botan::SHA_256::compress_digest_x86_avx2", [cli, "hash", "--algo=SHA-256", myfile], cpu=CPU.X86_64, masked_cpuid_bits=["intel_sha"]),
        StrubTest("Botan::SHA_256::compress_digest_x86_simd", [cli, "hash", "--algo=SHA-256", myfile], cpu=CPU.X86_64, masked_cpuid_bits=["intel_sha", "avx2"]),
    ]

    results = [test.run() for test in tests]
    os.remove(myfile)

    return 1 if any(not ok for ok in results) else 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

import sys
import datetime

# (C) 2018,2023,2025 Jack Lloyd
# Botan is released under the Simplified BSD License (see license.txt)

# Used to generate src/lib/math/mp/mp_monty_n.cpp

def monty_redc_code(n, p_dash1=False):

    lines = []

    fn_name = "bigint_monty_redc_pdash1" if p_dash1 else "bigint_monty_redc"

    if p_dash1:
        hdr = "void %s_%d(word r[%d], const word z[%d], const word p[%d], word ws[%d]) {\n" % (fn_name, n, n, 2*n, n, n)
    else:
        hdr = "void %s_%d(word r[%d], const word z[%d], const word p[%d], word p_dash, word ws[%d]) {\n" % (fn_name, n, n, 2*n, n, n)

    ftr = "\n}\n\n"

    lines.append("word3<word> accum;")
    lines.append("accum.add(z[0]);")
    if p_dash1:
        lines.append("ws[0] = accum.monty_step_pdash1();")
    else:
        lines.append("ws[0] = accum.monty_step(p[0], p_dash);")

    for i in range(1, n):
        for j in range(0, i):
            lines.append("accum.mul(ws[%d], p[%d]);" % (j, i-j))

        lines.append("accum.add(z[%d]);" % (i))
        if p_dash1:
            lines.append("ws[%d] = accum.monty_step_pdash1();" % (i))
        else:
            lines.append("ws[%d] = accum.monty_step(p[0], p_dash);" % (i))

    for i in range(0, n - 1):
        for j in range(i + 1, n):
            lines.append("accum.mul(ws[%d], p[%d]);" % (j, n + i-j))

        lines.append("accum.add(z[%d]);" % (n+i))
        lines.append("ws[%d] = accum.extract();" % (i))

    lines.append("accum.add(z[%d]);" % (2*n-1))

    lines.append("ws[%d] = accum.extract();" % (n - 1))
    lines.append("const word w1 = accum.extract();")

    lines.append("bigint_monty_maybe_sub<%d>(r, w1, ws, p);" % (n))

    return hdr + "\n".join(["   %s" % (line) for line in lines]) + ftr

def main(args = None):
    if args is None:
        args = sys.argv

    if len(args) <= 1:
        sizes = [4, 6, 8, 12, 16, 24, 32]
    else:
        sizes = map(int, args[1:])

    header = """/*
* This file was automatically generated by %s on %s
* All manual changes will be lost. Edit the script instead.
*
* Botan is released under the Simplified BSD License (see license.txt)
*/

#include <botan/internal/mp_core.h>

namespace Botan {

""" % (sys.argv[0], datetime.date.today().strftime("%Y-%m-%d"))

    footer = "}  // namespace Botan\n"

    with open('./src/lib/math/mp/mp_monty_n.cpp', encoding='utf8', mode='w') as monty_n:
        monty_n.write(header)

        for n in sizes:
            monty_n.write(monty_redc_code(n))

        monty_n.write(footer)

    return 0

if __name__ == '__main__':
    sys.exit(main())





#!/usr/bin/env python3

#
# Strips the KAT harness produced by the SPHINCS+ reference implementation down
# to a less space consuming version. This script was used to generate
# `src/tests/data/pubkey/sphincsplus.vec` test data from the *.rsp files of the
# reference implementation.
#
# (C) 2023 Jack Lloyd
# (C) 2023 René Meusel, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import sys
import hashlib
import binascii

class KatReader:
    def __init__(self, file):
        self.file = file
        self.last_mlen = None

    def next_value(self):
        while True:
            line = self.file.readline()

            if line == "":
                return (None, None)

            if line.startswith('#') or line == "\n":
                continue

            key, val = line.strip().split(' = ')

            return (key, val)

    def read_kats(self):
        kat = {}

        while True:
            key, val = self.next_value()

            if key is None:
                return # eof

            if key not in ['count', 'seed', 'mlen', 'msg', 'pk', 'sk', 'smlen', 'sm']:
                raise Exception("Unknown key %s" % (key))

            if key in ['count', 'mlen', 'smlen']:
                kat[key] = int(val)
            else:
                kat[key] = val

            if key == 'sm':
                yield kat
                kat = {}


def sha3_256(v):
    h = hashlib.sha3_256()
    h.update(v)
    return h.digest()


def sha256(v):
    h = hashlib.sha256()
    h.update(v)
    return h.digest()


def main(args = None):
    if args is None:
        args = sys.argv

    if len(args) < 3:
        print("Usage: %s <algo-spec as found in oids.txt> <*.rsp file> [optional: limit of KATs]")
        return 1

    param = args[1]
    katfile = args[2]
    limit = int(args[3]) if len(args) > 3 else 1

    reader = KatReader(open(katfile, encoding="utf-8"))

    hash_fn = sha256 if "sha2" in param.lower() else sha3_256

    cnt = 0
    for kat in reader.read_kats():
        if cnt >= limit:
            break
        cnt += 1

        # Remove the input message from the end of the 'sm' field
        signature = binascii.unhexlify(kat["sm"][:-kat["mlen"]*2])

        print("SphincsParameterSet = %s" % param)
        print("seed = %s" % kat["seed"])
        print("msg = %s" % kat["msg"])
        print("pk = %s" % kat["pk"])
        print("sk = %s" % kat["sk"])
        print("HashSig = %s" % binascii.hexlify(hash_fn(signature)).decode("utf-8").upper())
        print()

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
Runs all tests inmplemented in `ct_selftest`

(C) 2024 Jack Lloyd
    2024 Fabian Albert, René Meusel - Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import subprocess
import argparse
import os
import sys
from typing import Self
from enum import StrEnum, auto
import json

def run_command(cmd: list[str], is_text = True):
    """ Run the command . """
    return subprocess.run(cmd, capture_output=True, text=is_text, check=False)

def run_with_valgrind(cmd: list[str]):
    """ Run a command with valgrind. """
    valgrind_args = ['valgrind',
                     '-v',
                     '--error-exitcode=2']
    res = run_command(valgrind_args + cmd, is_text=False)
    # valgrind may output non-utf-8 characters
    res.stdout = res.stdout.decode("utf-8", errors="replace")
    res.stderr = res.stderr.decode("utf-8", errors="replace")
    return res

class ValgrindTest:
    """ A single test from ct_selftest """

    class Status(StrEnum):
        """ Defines the test result status """
        OK = auto()
        WARNING = auto()
        ERROR = auto()
        SKIP = auto()
        UNKNOWN = auto()

    def __init__(self, name, expect_failure, needs_special_config):
        self.name = name
        self.expect_failure = expect_failure
        self.needs_special_config = needs_special_config
        self.status = self.Status.UNKNOWN
        self.stdout = None
        self.stderr = None

    @staticmethod
    def from_line(line: str):
        info = line.split("\t")
        assert len(info) == 3
        return ValgrindTest(name=info[2],
                            expect_failure=info[0] == "true",
                            needs_special_config=info[1] == "true")

    def runnable(self, build_cfg):
        """ Decide whether or not to run this test given build config info """
        if not self.needs_special_config:
            return True

        if not build_cfg:
            return False  # test has special build requirements, but we have no info

        if self.name == "clang_vs_bare_metal_ct_mask":
            return build_cfg["cc_macro"] == "CLANG" and "-Os" in build_cfg["cc_compile_flags"]

        raise LookupError(f"Unknown special config test '{self.name}'")

    def run(self, exe_path: str, build_config):
        """ Run the test and return whether it succeeded """

        if not self.runnable(build_config):
            self.status = self.Status.SKIP
        else:
            result = run_with_valgrind([exe_path, self.name])
            self.stdout = result.stdout
            self.stderr = result.stderr

            if result.returncode == 1:
                # The ct_selftest program reported an error
                self.status = self.Status.WARNING
            else:
                failed = result.returncode != 0
                if failed != self.expect_failure:
                    # Valgrind reported an error (or not), but we expected the opposite
                    self.status = self.Status.ERROR
                else:
                    # Everything behaved as expected
                    self.status = self.Status.OK

        return self.status

    @staticmethod
    def read_test_list(ct_selftest_test_list: str) -> list[Self]:
        """ Read the list of tests from the output of `ct_selftest --list`. """

        return [ValgrindTest.from_line(line) for line in ct_selftest_test_list.split("\n")[2:] if line]


def main(): # pylint: disable=missing-function-docstring
    parser = argparse.ArgumentParser("ct_selftests")
    parser.add_argument("ct_selftest_path", help="Path to the ct_selftest executable")
    parser.add_argument("--build-config-path", help="Path to Botan's build-config.json file", default="")

    args = parser.parse_args()

    # Check if the path is valid
    if not os.path.isfile(args.ct_selftest_path):
        raise FileNotFoundError(f"Invalid path: {args.ct_selftest_path}")

    def find_test_list():
        test_list_result = run_command([args.ct_selftest_path, "--list"])
        if test_list_result.returncode != 0:
            raise RuntimeError("Failed to collect the test list from ct_selftest")
        return ValgrindTest.read_test_list(test_list_result.stdout)

    def open_build_config(build_config_path):
        if not build_config_path:
            return None

        if not os.path.isfile(build_config_path):
            raise FileNotFoundError(f"Invalid path: {build_config_path}")

        with open(build_config_path, encoding="utf-8") as build_info_file:
            return json.load(build_info_file)

    test_list = find_test_list()
    build_config = open_build_config(args.build_config_path)

    for test in test_list:
        print(f"running {test.name}... ", end="", flush=True)
        print(test.run(args.ct_selftest_path, build_config))
        if test.status not in [ValgrindTest.Status.OK, ValgrindTest.Status.SKIP]:
            if test.stdout:
                print("stdout:")
                print(test.stdout)
            if test.stderr:
                print("stderr:")
                print(test.stderr)


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2024 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)

Setup script for OSS-Fuzz
"""

import os
import sys
import subprocess

if len(sys.argv) != 3:
    print("Expected args: %s <cxx> <cxxflags>" % (sys.argv[0]))
    sys.exit(1)

cxx = sys.argv[1]
cxxflags = sys.argv[2]

args = [
    "--cc-bin=%s" % (cxx),
    "--cc-abi-flags=%s" % (cxxflags),
    "--without-documentation",
    "--disable-shared",
    "--disable-modules=locking_allocator",
    "--unsafe-fuzzer-mode",
    "--build-fuzzers=libfuzzer",
    "--build-targets=static",
    "--without-os-features=getrandom,getentropy",
    "--disable-modules=system_rng,processor_rng",
    "--with-fuzzer-lib=FuzzingEngine",
]

if cxxflags.find("-fsanitize=memory") > 0:
    args.append('--disable-asm')

configure = os.path.realpath(os.path.join(os.path.realpath(__file__), '..', '..', '..', 'configure.py'))

subprocess.run([configure] + args)



#!/usr/bin/env python3

"""
(C) 2023 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import re
import os

def main(args = None):
    if args is None:
        args = sys.argv

    if len(args) != 2:
        print("Usage: %s <lcov_filename>" % (args[0]))
        return 1

    header_re = re.compile(R'^SF:(.*/build/include/(public|internal|external)/botan(/internal)?/.*)')

    new_content = ""

    for line in open(args[1]):
        match = header_re.match(line)
        if match is not None:
            inc_path = match.group(1)

            if not os.path.islink(inc_path):
                print("ERROR %s in lcov but not a symlink" % (inc_path))
                return 1

            new_path = os.path.realpath(inc_path)
            print("Rewrote %s to %s in lcov file" % (inc_path, new_path))
            new_content += "SF:%s\n" % (new_path)
        else:
            new_content += line

    fd = open(args[1], 'w')
    fd.write(new_content)
    fd.close()

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2021,2025 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import re
import datetime
import os
import errno
from textwrap import dedent, indent
from jinja2 import Environment, FileSystemLoader
from addchain import addchain_code

"""
NOTE: This script requires the Jinja templating library to be installed.

This script generates the following files

src/lib/pubkey/ec_group/ec_named.cpp
src/lib/math/pcurves/pcurves_instance.h
src/lib/math/pcurves/pcurves.cpp

Additionally if a group is given in ec_groups.txt with an `Impl` that contains
"pcurves", and no pcurves implementation exists on disk, a default version will be
created. This step requires that addchain (https://github.com/mmcloughlin/addchain)
be installed.
"""

def curve_info(src):
    re_kv = re.compile('([A-Za-z]+) = ([0-9A-Za-z-_\\. ]+)')

    current = {}

    for line in src:
        line = line.strip()
        if line == '':
            continue

        kv = re_kv.match(line)
        if kv is None:
            raise Exception('Unknown line', line)

        key = kv.group(1)
        val = kv.group(2)

        if key in ['Name']:
            current[key] = val
        elif key in ['OID', 'Impl']:
            current[key] = val.split(' ')
        elif key in ['A']:
            if val == '-3':
                current[key] = current['P'] - 3
            else:
                current[key] = int(val, 16)
        elif key in ['P', 'B', 'X', 'Y', 'N']:
            current[key] = int(val, 16)

        if key == 'N':
            current["N32"] = current["N"] & 0xFFFFFFFF
            current["OIDExpr"] = ['OID{%s}' % (oid.replace('.', ', ')) for oid in current['OID']]
            yield current
            current = {}

def format_names(names):
    # This would be quite complicated to render in the Jinja template language so
    # we pre-render it as a string and insert it directly
    legacy = []
    generic = []
    pcurves = []
    pcurves_no_generic = []

    for (nm, impl) in names:
        if 'pcurve' in impl:
            pcurves.append(nm)
            if 'generic' not in impl:
                pcurves_no_generic.append(nm)
        elif 'generic' in impl:
            generic.append(nm)
        else:
            assert 'legacy' in impl
            legacy.append(nm)

    legacy_macro = "defined(BOTAN_HAS_LEGACY_EC_POINT)"
    generic_macro = "defined(BOTAN_HAS_PCURVES_GENERIC)"
    for nm in sorted(pcurves):
        nm_macro = "BOTAN_HAS_PCURVES_%s" % (nm.upper())

        if nm in pcurves_no_generic:
            yield "#if defined(%s) || %s" % (nm_macro, legacy_macro)
            yield "      // Not supported by pcurves_generic"
        else:
            yield "#if defined(%s) || %s || %s" % (nm_macro, legacy_macro, generic_macro)
        yield '      \"%s\",' % (nm)
        yield "#endif\n"

    yield "#if %s || %s" % (legacy_macro, generic_macro)
    for nm in sorted(generic):
        yield '      \"%s\",' % (nm)
    yield "#endif\n"

    yield "#if %s" % (legacy_macro)
    for nm in sorted(legacy):
        yield '      \"%s\",' % (nm)
    yield "#endif\n"

def datestamp():
    current_date = datetime.datetime.now()
    return int(current_date.strftime("%Y%m%d"))

class OmitFirstLine:
    def __init__(self):
        self.first_line = True

    def __call__(self, line):
        r = not self.first_line
        self.first_line = False
        return r

def main():
    curves = [c for c in curve_info(open('./src/build-data/ec_groups.txt'))]

    pcurves = []
    for c in curves:
        if 'pcurve' in c['Impl']:
            pcurves.append(c)

    this_script = sys.argv[0]
    date = datetime.date.today().strftime("%Y-%m-%d")

    env = Environment(loader=FileSystemLoader("src/build-data/templates"))

    # write ec_named.cpp
    with open('./src/lib/pubkey/ec_group/ec_named.cpp', encoding='utf8', mode='w') as ec_named:
        template = env.get_template("ec_named.cpp.in")

        named_groups = "\n".join(format_names([(c['Name'], c['Impl']) for c in curves]))

        ec_named.write(template.render(script=this_script, date=date, curves=curves, named_groups=named_groups.strip()))
        ec_named.write("\n")

    # write pcurves_instance.h
    with open('./src/lib/math/pcurves/pcurves_instance.h', encoding='utf8', mode='w') as pcurves_h:
        template = env.get_template("pcurves_instance.h.in")
        pcurves_h.write(template.render(script=this_script, date=date, pcurves=pcurves))
        pcurves_h.write("\n")

    # write pcurves.cpp
    with open('./src/lib/math/pcurves/pcurves.cpp', encoding='utf8', mode='w') as pcurves_cpp:
        template = env.get_template("pcurves.cpp.in")
        pcurves_cpp.write(template.render(script=this_script, date=date, pcurves=pcurves))
        pcurves_cpp.write("\n")

    # Check if any pcurves modules need a new stub impl
    for pcurve in pcurves:
        curve = pcurve["Name"]
        mod_dir = './src/lib/math/pcurves/pcurves_%s' % (curve)
        info_path = os.path.join(mod_dir, 'info.txt')
        impl_path = os.path.join(mod_dir, f'pcurves_{curve}.cpp')

        if os.access(impl_path, os.R_OK):
            continue

        addchain_fe2 = addchain_code(pcurve['P'] - 3, 0)
        addchain_fe_sqrt = addchain_code((pcurve['P'] + 1) // 4, 0) if pcurve['P'] % 4 == 3 else None
        addchain_scalar = addchain_code(pcurve['N'] - 2, 0)

        try:
            os.makedirs(mod_dir)
        except OSError as ex:
            if ex.errno != errno.EEXIST:
                raise

        module_define = f"PCURVES_{curve.upper()}"
        if not re.match('^[0-9A-Za-z_]{3,30}$', module_define):
            raise ValueError(f"Invalid preprocessor define name ({module_define}) for new pcurve module")

        with open(info_path, 'w', encoding='utf8') as info_file:
            info_file.write(dedent(f"""\
                <internal_defines>
                {module_define} -> {datestamp()}
                </internal_defines>

                <module_info>
                name -> "PCurve {curve}"
                </module_info>

                <requires>
                pcurves_impl
                </requires>
            """))

        with open(impl_path, 'w', encoding='utf8') as src_file:
            crandall = (1 << pcurve["P"].bit_length()) - pcurve["P"]
            if crandall > 2**32:
                crandall = 0

            template = env.get_template("pcurves_stub.cpp.in")
            src_file.write(template.render(curve = pcurve,
                                           crandall=crandall,
                                           addchain_fe2=indent(addchain_fe2, 9 * ' ', OmitFirstLine()),
                                           addchain_fe_sqrt=indent(addchain_fe_sqrt, 9 * ' ', OmitFirstLine()) if addchain_fe_sqrt else None,
                                           addchain_scalar=indent(addchain_scalar, 9 * ' ', OmitFirstLine())))
            src_file.write("\n")


    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
This is a wrapper around the XMSS reference implementation CLI tools
to generate test data readable py test_xmss.cpp

Uses the XMSS reference implementation from here:
   https://github.com/XMSS/xmss-reference

(C) 2023 René Meusel (Rohde & Schwarz Cybersecurity)

Botan is released under the Simplified BSD License (see license.txt)
"""

import asyncio
import tempfile
import sys
import binascii


async def run(cmd, args):
    vc = await asyncio.create_subprocess_exec(cmd, *args,
                                              stdout=asyncio.subprocess.PIPE,
                                              stderr=asyncio.subprocess.PIPE)
    (stdout, stderr) = await vc.communicate()
    if stderr:
        raise RuntimeError(
            "Process execution failed with: " + stderr.decode("utf-8"))
    return stdout


def tohex(b):
    return binascii.hexlify(b).decode("utf-8")


def fromhex(s):
    return binascii.unhexlify(s)


KEYGEN_UTIL = "ui/xmss_keypair"
SIGN_UTIL = "ui/xmss_sign"

TEST_MESSAGES = [b'', fromhex("01020304"), fromhex("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")]


class Keypair:
    @staticmethod
    async def generate(xmss_param):
        for _ in range(5):
            keypair = await run(KEYGEN_UTIL, [xmss_param])
            if len(keypair) < 8:
                continue

            oid = keypair[0:4]

            # Slicing fails if the oid appears in the public key by chance.
            # This is just a simple test generation script, so we don't really
            # care about the bias this introduces.
            if keypair.count(oid) == 2:
                return Keypair(xmss_param, keypair)

        raise RuntimeError("Keypair generation failed")

    def __init__(self, xmss_params, keypair):
        self.keypair = keypair
        self.oid = self.keypair[0:4]
        self.params_name = xmss_params[5:]

        # transform the public/private keys into the format Botan expects
        sk_slice = self.keypair.find(self.oid, 4)
        self.public_key = self.keypair[:sk_slice]
        secret_portion = self.keypair[sk_slice+4:-len(self.public_key)+4]
        n = int((len(secret_portion) - 4) / 2)
        self.idx = secret_portion[0:4]
        self.secret_seed = secret_portion[4:n+4]
        self.secret_prf = secret_portion[n+4:]
        self.public_seed = self.public_key[n+4:n*2+4]
        wots_derivation_method = "\x02" # WOTS+ derivation as described in NIST SP.800-208
        self.secret_key = self.public_key + self.idx + self.secret_prf + self.secret_seed + wots_derivation_method

    async def sign(self, message):
        with tempfile.NamedTemporaryFile() as msg_file, tempfile.NamedTemporaryFile() as key_file:
            key_file.write(self.keypair)
            key_file.flush()
            msg_file.write(message)
            msg_file.flush()
            signature_and_msg = await run(SIGN_UTIL, [key_file.name, msg_file.name])
            return signature_and_msg[:-len(message)] if len(message) != 0 else signature_and_msg


async def join(strings):
    return "\n".join(await asyncio.gather(*strings))


async def make_xmss_sig_vec_entry(xmss_param):
    async def entry(msg):
        kp = await Keypair.generate(xmss_param)
        out = "Params = {}\n".format(kp.params_name)
        out += "Msg = {}\n".format(tohex(msg))
        out += "PrivateKey = {}\n".format(tohex(kp.secret_key))
        sig = await kp.sign(msg)
        out += "Signature = {}\n".format(tohex(sig))
        return out

    return await join([entry(msg) for msg in TEST_MESSAGES])


async def make_xmss_verify_vec_entry(xmss_param):
    async def entry(msg):
        kp = await Keypair.generate(xmss_param)
        out = "Params = {}\n".format(kp.params_name)
        out += "Msg = {}\n".format(tohex(msg))
        out += "PublicKey = {}\n".format(tohex(kp.public_key))
        sig = await kp.sign(msg)
        out += "Signature = {}\n".format(tohex(sig))
        return out

    return await join([entry(msg) for msg in TEST_MESSAGES])


async def make_xmss_keygen_vec_entry(xmss_param):
    kp = await Keypair.generate(xmss_param)
    out = "Params = {}\n".format(xmss_param)
    out += "SecretSeed = {}\n".format(tohex(kp.secret_seed))
    out += "PublicSeed = {}\n".format(tohex(kp.public_seed))
    out += "SecretPrf = {}\n".format(tohex(kp.secret_prf))
    out += "PublicKey = {}\n".format(tohex(kp.public_key))
    out += "PrivateKey = {}\n".format(tohex(kp.secret_key))
    return out


if len(sys.argv) < 3:
    print("Usage: {} <test vector: 'sig', 'verify', 'keygen'> <XMSS Parameter Set Name(s)>".format(
        sys.argv[0]))
    sys.exit(1)

funs = {
    "sig": make_xmss_sig_vec_entry,
    "verify": make_xmss_verify_vec_entry,
    "keygen": make_xmss_keygen_vec_entry
}

tv = sys.argv[1]
if tv not in funs:
    print("Unknown test vector: %s" % tv)
    sys.exit(1)


async def main():
    print(await join([funs[tv](p) for p in sys.argv[2:]]))

asyncio.run(main())



#!/usr/bin/env python3
# coding=utf8

"""
Botan CI check headers script

(C) 2023 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import os
import json
import re

def main(args=None):
    if args is None:
        args = sys.argv

    if len(args) < 2:
        print("Usage: %s <build_config.json>" % args[0])
        return 1

    with open(os.path.join(args[1]), encoding='utf8') as f:
        build_config = json.load(f)

        public_include_dir = build_config['public_include_path']
        internal_include_dir = build_config['internal_include_path']

        internal_inc = re.compile(r'#include <botan/(internal/.*)>')

        for header in build_config['public_headers']:
            contents = open(os.path.join(public_include_dir, header), encoding='utf8').read()

            match = internal_inc.search(contents)
            if match:
                print("ERROR: Public header '%s' includes an internal header '%s'" % (header, match.group(1)))
                return 1


        all_headers = build_config['public_headers'] + build_config['internal_headers']
        header_guard = re.compile(r'\n#define BOTAN_([A-Z0-9_]{3,})_H_\n')
        header_guards = {}

        for header in all_headers:

            if header in build_config['public_headers']:
                path = os.path.join(public_include_dir, header)
            else:
                path = os.path.join(internal_include_dir, header)

            contents = open(path, encoding='utf8').read()

            match = header_guard.search(contents)
            if not match:
                print("ERROR: Header '%s' is missing an appropriate header guard" % (header))
                return 1

            guard = match.group(1)
            if guard in header_guards:
                print("ERROR: Duplicate header guard (%s) in '%s' and '%s'" % (guard, header, header_guards[guard]))
                return 1
            else:
                header_guards[guard] = header

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2024 Jack Lloyd
(C) 2024 René Meusel - Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

# Use this as a script to read the repository's configuration file and access
# its contents. For instance:
#
#   $ python3 repo_config.py all
#   $ python3 repo_config.py get BORINGSSL_REPO
#
# This might also be used as a module to access the configuration file from
# other python scripts.

import argparse
import os
import re
import sys

_DEFAULT_REPO_CONFIG_LOCATION = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'configs', 'repo_config.env'))

class RepoConfig(dict):
    """ Reads the repository's configuration file and provides access to its variables. """

    def __init__(self, env_file: str = _DEFAULT_REPO_CONFIG_LOCATION):
        self._file_path = env_file
        parser = re.compile(r'^(?P<var>\w+)\s*=\s*((?P<val>[^\"]\S+)|\"(?P<quoted_val>\S+)\")\s*$')
        with open(self._file_path, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                if m := parser.match(line):
                    var = m.group('var')
                    val = m.group('val') or m.group('quoted_val')
                    try:
                        self[var] = int(val)
                    except ValueError:
                        self[var] = val

    @property
    def config_file_path(self):
        return self._file_path

def main():
    def print_all(cfg: RepoConfig, *_):
        print('\n'.join(f'{key}={value}' for key, value in cfg.items()))

    def list_vars(cfg: RepoConfig, *_):
        print('\n'.join(key for key in cfg))

    def get_var(cfg: RepoConfig, args):
        if args.var not in cfg:
            print(f'Variable "{args.var}" not found in the configuration file.', file=sys.stderr)
            raise KeyError()
        print(cfg[args.var])

    def print_config_file_path(cfg, *_):
        print(cfg.config_file_path)

    parser = argparse.ArgumentParser(description='Read and process a .env file.')
    subparsers = parser.add_subparsers(dest='command', required=True)

    parser_all = subparsers.add_parser('all', help='Print all variables and their values.')
    parser_all.set_defaults(dispatch=print_all)

    parser_all = subparsers.add_parser('list', help='Print all variable names.')
    parser_all.set_defaults(dispatch=list_vars)

    parser_get = subparsers.add_parser('get', help='Get the value of a specific variable.')
    parser_get.add_argument('var', type=str, help='The variable name to retrieve.')
    parser_get.set_defaults(dispatch=get_var)

    parser_file = subparsers.add_parser('file', help='Print the path to the configuration file.')
    parser_file.set_defaults(dispatch=print_config_file_path)

    args = parser.parse_args()
    try:
        args.dispatch(RepoConfig(), args)
    except Exception:
        return 1

    return 0

if __name__ == '__main__':
    sys.exit(main())



import os
import sys
import subprocess
import re
import multiprocessing

class BuildError(Exception):
    pass


_MAKEFILE = "Makefile"


def get_concurrency():
    def_concurrency = 2
    max_concurrency = 16

    try:
        return min(max_concurrency, multiprocessing.cpu_count())
    except ImportError:
        return def_concurrency


def run_cmd(cmd, workingdir=None):
    if isinstance(cmd, str):
        print('> running: ' + cmd)
        shell = True
    else:
        print('> running: ' + ' '.join(cmd))
        shell = False
    sys.stdout.flush()

    try:
        subprocess.run(cmd, shell=shell, check=True, cwd=workingdir)
    except subprocess.CalledProcessError as ex:
        raise BuildError('External command failed, aborting...') from ex


def _find_regex_in_makefile(regex):
    if not os.path.exists(_MAKEFILE):
        raise BuildError('No Makefile found. Maybe run ./configure.py?')

    with open(_MAKEFILE, 'r', encoding="utf-8") as f:
        return re.search(regex, f.read())


def get_test_binary_name():
    match = _find_regex_in_makefile(r'TEST\s*=\s*([^\n]+)\n')
    if not match:
        raise BuildError('Test binary name not found in Makefile')
    test_file = os.path.split(match.group(1))[1]
    if not test_file:
        raise BuildError(
            'Cannot make sense of test binary name: ' + match.group(0))

    return test_file



#!/usr/bin/env python3

# (C) 2024 Jack Lloyd
#
# Botan is released under the Simplified BSD License (see license.txt)

"""
Runs https://github.com/mmcloughlin/addchain and formats
the code suitable for pcurves
"""

import sys
import subprocess

def addchain_gen(n):
    search = subprocess.Popen(['addchain', 'search', str(n)],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

    (stdout, stderr) = search.communicate()

    gen = subprocess.Popen(['addchain', 'gen'],
                           stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)

    gen.stdin.write(stdout)

    (stdout, stderr) = gen.communicate()

    return stdout.decode('utf8').split('\n')

def addchain_code(n, indent=3):
    vars = set([])

    output = []

    for line in addchain_gen(n):
        if line == '':
            continue
        c = line.strip().split()

        if c[0] == 'tmp':
            continue

        decl = '' if c[1] in vars else 'auto '

        if c[0] == 'double':
            assert(len(c) == 3)
            output.append("%s%s = %s.square()" % (decl, c[1], c[2]))
            vars.add(c[1])
        elif c[0] == 'add':
            assert(len(c) == 4)
            if c[1] == c[2]:
                output.append("%s *= %s" % (c[1], c[3]))
            elif c[1] == c[3]:
                output.append("%s *= %s" % (c[1], c[2]))
            else:
                if c[2] < c[3]:
                    output.append("%s%s = %s * %s" % (decl, c[1], c[2], c[3]))
                else:
                    output.append("%s%s = %s * %s" % (decl, c[1], c[3], c[2]))
            vars.add(c[1])
        elif c[0] == 'shift':

            assert(len(c) == 4)
            if c[1] != c[2]:
                output.append("%s%s = %s" % (decl, c[1], c[2]))
            output.append("%s.square_n(%s)" % (c[1], c[3]))
            vars.add(c[1])
        else:
            raise Exception("Don't know what to do with %s" % (c[0]))

    output.append('return z')

    ws = " " * indent
    return '\n'.join(['%s%s;' % (ws, line) for line in output])

def main(args = None):
    if args is None:
        args = sys.argv
    if len(args) != 2:
        print("Usage: %s <n>" % (args[0]))
        return 1

    n = int(args[1], 0)
    print(addchain_code(n))
    return 0

if __name__ == '__main__':
    sys.exit(main())



# TLS-Anvil calls a trigger script before each test handshake to make the
# client-under-test initiate a new TLS connection. This server listens for
# HTTP GET /trigger requests and spawns a fresh Botan tls_client process
# for each one.
#
# (C) 2026 Jack Lloyd
#
# Botan is released under the Simplified BSD License (see license.txt)

import sys
import os
import signal
import subprocess
import argparse
import threading
from http.server import HTTPServer, BaseHTTPRequestHandler


# Populated from command-line arguments in main()
botan_cli = None
tls_anvil_policy = None
tls_anvil_host = None
tls_anvil_port = None
client_log_dir = None
client_timeout = 30

# Track spawned client processes so we can clean them up
client_processes = []
trigger_count = 0


def reap_client(proc):
    """Wait for the timeout, then kill the process if still running.

    While waiting, periodically write to stdin so that tls_client calls
    send() and flushes any deferred KeyUpdate response."""
    import time
    deadline = time.monotonic() + client_timeout
    while time.monotonic() < deadline:
        if proc.poll() is not None:
            break
        if proc.stdin:
            try:
                proc.stdin.write(b".\n")
                proc.stdin.flush()
            except OSError:
                break
        try:
            proc.wait(timeout=1)
            break
        except subprocess.TimeoutExpired:
            pass
    if proc.poll() is None:
        proc.kill()
        proc.wait()
    if proc.stdin:
        try:
            proc.stdin.close()
        except OSError:
            pass


def cleanup_finished_clients():
    """Remove completed client processes from the tracking list."""
    still_running = []
    for proc in client_processes:
        if proc.poll() is None:
            still_running.append(proc)
    client_processes[:] = still_running


def kill_all_clients():
    """Terminate all tracked client processes."""
    for proc in client_processes:
        try:
            proc.kill()
            proc.wait()
            if proc.stdin:
                proc.stdin.close()
        except OSError:
            pass
    client_processes.clear()


class TriggerHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path == "/trigger":
            global trigger_count
            cleanup_finished_clients()

            log_file = None
            if client_log_dir:
                log_path = os.path.join(client_log_dir, f"client_{trigger_count}.log")
                log_file = open(log_path, 'w', encoding='utf-8')

            trigger_count += 1

            proc = subprocess.Popen(
                [botan_cli, "tls_client", tls_anvil_host,
                 f"--port={tls_anvil_port}",
                 f"--policy={tls_anvil_policy}",
                 "--ignore-cert-error"],
                stdin=subprocess.PIPE,
                stdout=log_file or subprocess.DEVNULL,
                stderr=log_file or subprocess.DEVNULL,
            )
            client_processes.append(proc)

            # Close the log file in this process; the child has its own fd
            if log_file:
                log_file.close()

            # Start a reaper thread that kills the process after the timeout
            threading.Thread(target=reap_client, args=(proc,), daemon=True).start()

            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"OK\n")

        elif self.path == "/shutdown":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"OK\n")

            def shutdown():
                self.server.shutdown()
            threading.Thread(target=shutdown, daemon=True).start()

        else:
            self.send_response(404)
            self.end_headers()

    def log_message(self, format, *args):
        # Suppress request logging noise
        pass


def main(args=None):
    global botan_cli, tls_anvil_policy, tls_anvil_host, tls_anvil_port, client_log_dir, client_timeout

    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(description="TLS-Anvil trigger server for Botan client testing")
    parser.add_argument("--botan-cli", required=True, help="Path to botan executable")
    parser.add_argument("--tls-anvil-policy", default="default", help="Policy to use for TLS-Anvil testing")
    parser.add_argument("--tls-anvil-host", default="127.0.0.1", help="Host where TLS-Anvil listens")
    parser.add_argument("--tls-anvil-port", type=int, required=True, help="Port where TLS-Anvil listens")
    parser.add_argument("--listen-port", type=int, default=8090, help="Port for this trigger server")
    parser.add_argument("--log-dir", default=None, help="Directory for client log files")
    parser.add_argument("--timeout", type=int, default=10,
                        help="Kill client processes after this many seconds (default: %(default)s)")

    parsed = parser.parse_args(args)

    botan_cli = parsed.botan_cli
    tls_anvil_policy = parsed.tls_anvil_policy
    tls_anvil_host = parsed.tls_anvil_host
    tls_anvil_port = parsed.tls_anvil_port
    client_log_dir = parsed.log_dir
    client_timeout = parsed.timeout

    if client_log_dir:
        os.makedirs(client_log_dir, exist_ok=True)

    server = HTTPServer(("127.0.0.1", parsed.listen_port), TriggerHandler)
    print(f"Trigger server listening on 127.0.0.1:{parsed.listen_port}", flush=True)

    def handle_signal(signum, frame):
        kill_all_clients()
        sys.exit(0)

    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)

    try:
        server.serve_forever()
    finally:
        kill_all_clients()
        server.server_close()


if __name__ == "__main__":
    main()
